from typing import Mapping
from cadCAD_tools.execution import easy_run
from pandas import DataFrame
from cadCAD_tools.types import Param, SystemParameters, InitialState, List, ParamSweep, TimestepBlock
from itertools import product
from dataclasses import dataclass


def sweep_cartesian_product(sweep_params: Mapping[str, tuple]) -> Mapping[str, tuple]:
    """
    Makes a cartesian product from dictionary values.
    This is useful for plugging inside the sys_params dict, like:
    ```python
    sweep_params = {'a': [0.1, 0.2], 'b': [1, 2]}
    product_sweep 
    sys_params = {**cartesian_product_sweep(sweep_params),
                  'c': [0.1]}
    ```
    Usage:
    >>> sweep_params = {'a': [0.1, 0.2], 'b': [1, 2]}
    >>> cartesian_product_sweep(sweep_params)
    {'a': [0.1, 0.1, 0.2, 0.2], 'b': [1, 2, 1, 2]}
    """
    cartesian_product = product(*sweep_params.values())
    transpose_cartesian_product = zip(*cartesian_product)
    zipped_sweep_params = zip(sweep_params.keys(), transpose_cartesian_product)
    sweep_dict = dict(zipped_sweep_params)
    sweep_dict = {k: tuple(v) for k, v in sweep_dict.items()}
    return sweep_dict


def prepare_params(params: SystemParameters,
                   cartesian_sweep: bool = False) -> Mapping[str, List[object]]:
    simple_params = {k: [v.value]
                     for k, v in params.items()
                     if type(v) is Param}

    sweep_params: Mapping = {k: v.value
                             for k, v in params.items()
                             if type(v) is ParamSweep}
    if cartesian_sweep is True:
        sweep_params = sweep_cartesian_product(sweep_params)
    else:
        pass

    cleaned_params = {**simple_params, **sweep_params}
    return cleaned_params


def prepare_state(state: InitialState) -> Mapping[str, object]:
    cleaned_state = {k: v.value
                     for k, v in state.items()}
    return cleaned_state


@dataclass
class ConfigurationWrapper():
    initial_state: InitialState
    params: SystemParameters
    timestep_block: TimestepBlock
    timesteps: int
    samples: int

    def run(self, *args, **kwargs) -> DataFrame:
        output = easy_run(prepare_state(self.initial_state),
                          prepare_params(self.params),
                          self.timestep_block,
                          self.timesteps,
                          self.samples,
                          *args,
                          **kwargs)
        return output


from cadCAD_tools.types import Signal, StateUpdate, VariableUpdate

def generic_suf(variable: str,
                signal: str='') -> StateUpdate:
    """
    Generate a State Update Function that assigns the signal value to the 
    given variable. By default, the signal has the same identifier as the
    variable.
    """
    if signal is '':
        signal = variable
    else:
        pass

    def suf(_1, _2, _3, _4, signals: Signal) -> VariableUpdate:
        return (variable, signals[signal])
    return suf


from setuptools import setup, find_packages

long_description = """
cadCAD tools
"""

setup(name='cadCAD_tools',
      version='0.0.1.7',
      description="tools",
      long_description=long_description,
      url='https://github.com/cadCAD-org/cadCAD-tools',
      author='Danilo Lessa Bernardineli',
      author_email='danilo@block.science, danilo.lessa@gmail.com',
      packages=find_packages(),
      install_requires=['pandas', 'tqdm', 'numpy', 'plotly']
)



from tqdm.auto import tqdm
import pandas as pd
import plotly.express as px
import numpy as np


def visualize_elapsed_time_per_ts(df: pd.DataFrame, relative=False) -> None:
    indexes = ['simulation', 'run', 'timestep', 'substep']

    z_df = df.set_index(indexes)
    first_time = z_df.query(
        'timestep == 1 & substep == 1').reset_index([-1, -2]).run_time
    s = (z_df.run_time - first_time)
    s.name = 'time_since_start'

    z_df = z_df.join(s)
    s = z_df.groupby(indexes[:-1]).time_since_start.max()

    fig_df = s.reset_index()
    if relative is True:
        s = fig_df.groupby(indexes[:-2]).time_since_start.diff()
        s.name = 'psub_duration'
        fig_df = fig_df.join(s)

        y_col = 'psub_duration'
    else:
        y_col = 'time_since_start'
        
    fig = px.box(fig_df,
                x='timestep',
                y=y_col)

    return fig


def visualize_substep_impact(df: pd.DataFrame, relative=True, **kwargs) -> None:
    indexes = ['simulation', 'run', 'timestep', 'substep']

    new_df = df.copy()
    new_df = new_df.assign(psub_time=np.nan).set_index(indexes)

    # Calculate the run time associated with PSUBs
    for ind, gg_df in tqdm(df.query('substep > 0').groupby(indexes[:-1])):
        g_df = gg_df.reset_index()
        N_rows = len(g_df)
        substep_rows = list(range(N_rows))[1:-1:2]

        for substep_row in substep_rows:
            t1 = g_df.run_time[substep_row - 1]
            t2 = g_df.run_time[substep_row + 1]
            dt = t2 - t1
            g_df.loc[substep_row, 'psub_time'] = dt
        g_df = g_df.set_index(indexes)
        new_df.loc[g_df.index, 'psub_time'] = g_df.psub_time

    fig_df = new_df.reset_index().dropna(subset=['psub_time'])


    if 'substep_label' in fig_df.columns:
        x_col = 'substep_label'
    else:
        x_col = 'substep'
        fig_df[x_col] = fig_df[x_col] / 2

    if relative is True:
        fig_df = fig_df.assign(relative_psub_time=fig_df.groupby(indexes[:-1]).psub_time.apply(lambda x: x / x.sum()))
        y_col = 'relative_psub_time'
    else:
        y_col = 'psub_time'

    inds = fig_df[y_col] < fig_df[y_col].quantile(0.95)
    inds &= fig_df[y_col] > fig_df[y_col].quantile(0.05)

    fig = px.box(fig_df[inds],
                 x=x_col,
                 y=y_col,
                 **kwargs)

    return fig
